/* 
   sitecopy, remote web site copy manager.
   Copyright (C) 1998, Joe Orton <joe@orton.demon.co.uk>
                                                                     
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
  
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
  
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <ctype.h>
#include <string.h>
#include <unistd.h>
#include <time.h>

#include "sites.h"
#include "ftp.h"
#include "sitecopy.h"

/* This is the string used in the storage files to indicate the
 * line is a directory rather than a file */
#define DIRWORD "dir"

/* Groovy macro to remove items from a doubly linked list...
 * give it the item to delete and the list head pointer */
#define LL_DELETE(item,head)			\
if( item->prev ) {				\
    item->prev->next = item->next;		\
} else {					\
    head = item->next;				\
}						\
if( item->next ) {				\
    item->next->prev = item->prev;		\
}

/* From sitecopy.c */
extern enum debug_types debug_level;
extern char *progname;
extern char *copypath;

extern bool prompting;

/* From ftp.c */
extern char *ftp_error;

/* Temporary for use in ftw callback */
struct site_t *cur_site;

/* Functions for dealing with sites */

/* site_readsites will read ~/.sitecopyrc and fill the sites list */
int site_readsites( char *filename, struct site_t **sites ) {
    FILE *fp; int state, alpha;
    char buf[BUFSIZ];
    char *ch, *ptr, *key, *val;
    /* Holders for the site info */
    struct site_t *this_site, *last_site;

    if( (fp = fopen( filename, "r" )) == NULL ) {
	perror("fopen");
	return 1;
    } 

    key = malloc( BUFSIZ );
    val = malloc( BUFSIZ );

    last_site = this_site = NULL;

    while( fgets( buf, sizeof(buf), fp ) != NULL ) {
	state = 0;
	ptr = key;
	*key = *val = '\0';
	for( ch=buf; *ch!='\0' && *ch!='#'; ch++ ) {
	    alpha = !isspace(*ch);
	    switch( state ) {
	    case 0: 
		if(alpha) {
		    *(ptr++) = *ch;
		    state = 1;
		}
		break;
	    case 1:
		if(!alpha) {
		    *ptr = '\0';
		    ptr = val;
		    state = 2;
		} else {
		    *(ptr++) = *ch;
		}
		break;
	    case 2:
		if(alpha) {
		    *(ptr++) = *ch;
		    state = 3;
		}
		break;
	    case 3:
		if(!alpha) {
		    *ptr = '\0';
		    state = 4;
		} else {
		    *(ptr++) = *ch;
		}
	    default: /* ignores everything past end of value */
		break;
	    }
	}
	/* printf("Key [%s] Value: [%s]\n", key, val); */
	if( strlen( key ) > 0 ) {
	    if( strcmp( key, "site" ) == 0 ) {
		/* Beginning of a new Site */
		last_site = this_site;
		/* Allocate new item */
		this_site = malloc( sizeof( struct site_t ) );
		/* Zero it out */
		memset( this_site, 0, sizeof( struct site_t ) );
		if( last_site != NULL ) { /* next site */
		    last_site->next = this_site;
		} else { /* First site */
		    *sites = this_site;
		}
		this_site->name = strdup( val );
		this_site->files = NULL;
		/* defaults... */
		this_site->perms = sitep_ignore;
		this_site->nodelete = false;
		this_site->checkmoved = false;
		/* inits... 
		this_site->nummoved = 0;
		this_site->numdeleted = 0;
		this_site->numchanged = 0;
		this_site->numnew = 0; */
		/* Now work out the info filename */
		this_site->infofile = malloc( strlen(copypath) + strlen(val) + 1);
		strcpy( this_site->infofile, copypath );
		strcat( this_site->infofile, val );
	    } else if( this_site == NULL ) {
		/* TODO: Only do this if we have a valid keyword */
		fprintf( stderr, "Error: %s before `site' keyword\n", key );
	    } else if( strcmp( key, "username" ) == 0 ) {
		/* username */
		this_site->username = strdup( val );
	    } else if( strcmp( key, "server" ) == 0 ) {
		if( val ) this_site->server = strdup( val );
	    } else if( strcmp( key, "password" ) == 0 ) {
		if( val ) this_site->password = strdup( val );
	    } else if( strcmp( key, "remote" ) == 0 ) {
		if( strlen(val) > 0 ) {
		    /* Relative filenames must start with "~/" */
		    if( val[0] == '~' ) {
			if( strlen(val) > 1 ) {
			    if( val[1] == '/' ) {
				if( val[strlen(val)-1] != '/' )
				    strcat( val, "/" );
				this_site->remote_dir = strdup( val + 2 );
			    } else {
				fprintf( stderr, "Error: If you wish the remote directory to be relative to the login directory,\nprefix it with '~/'" );
				return 1;
			    }
			} else {
			    fprintf( stderr, "Error: You must include a valid remote directory, eg: '~/html' or '~/'." );
			    return 1;
			}
		    } else { /* Dirname doesn't begin with a ~ */
			if( val[strlen(val)-1] != '/' )
			    strcat( val, "/" );
			this_site->remote_dir = strdup( val );
		    }
		} else { /* strlen(val) == 0 */
		    fprintf(stderr,"Error: Remote directory not specified.\n");
		    return 1;
		}
	    } else if( strcmp( key, "local" ) == 0 ) {
		if( val[strlen(val)-1] != '/' )
		    strcat( val, "/" );
		this_site->local_dir = strdup( val );
	    } else if( strcmp( key, "permissions") == 0 ) {
		if( strcmp( val, "ignore" ) == 0 ) {
		    this_site->perms = sitep_ignore;
		} else if( strcmp( val, "exec" ) == 0 ) {
		    this_site->perms = sitep_exec;
		} else if( strcmp( val, "all" ) == 0 ) {
		    this_site->perms = sitep_all;
		} else {
		    fprintf( stderr, "Error: Unknown permissions mode %s!\n", val );
		    return 1;
		}
	    } else if( strcmp( key, "nodelete" ) == 0 ) {
		this_site->nodelete = true;
	    } else if( strcmp( key, "checkmoved" ) == 0 ) {
		this_site->checkmoved = true;
	    } else {
		fprintf( stderr, "Error: Unknown keyword %s!\n", key );
		return 1;
	    }
	}
    }
    free(key);
    free(val);
    fclose( fp );
    return 0;
}

int site_find( struct site_t **sites, struct site_t **the_site, char *sitename ) {
    struct site_t *current;
    if( sites == NULL) return 1; /* be sane */
    for( current = *sites; current!=NULL; current=current->next ) {
	if( strcmp( current->name, sitename ) == 0 ) {
	    /* We found it */
	    *the_site = current;
	    return site_verify( current );
	}
    }
    return 1;
}

int site_verify( struct site_t *any_site ) {
    if( any_site->name == NULL ) {
	fprintf( stderr, "Site name not specified.\n" );
    } else if( any_site->server == NULL ) {
	fprintf( stderr, "%s: Server not specified in site `%s'.\n", progname, any_site->name );
    } else if( any_site->username == NULL ) {
	fprintf( stderr, "%s: Username not specified in site `%s'.\n", progname, any_site->name );
    } else if( any_site->password == NULL ) {
	fprintf( stderr, "%s: Password not specified in site `%s'.\n", progname, any_site->name );
    } else if( any_site->remote_dir == NULL ) {
	fprintf( stderr, "%s: Remote directory not specified in site `%s'.\n", progname, any_site->name );
    } else if( any_site->local_dir == NULL ) {
	fprintf( stderr, "%s: Local directory not specified in site `%s'.\n", progname, any_site->name );
    } else {
	return 0;
    }
    return 1;
}

/* Updates the remote files list of the given site.
 * Tricky, this - have to do a load of LIST's, and try and
 * parse the response sensibly */
int site_resync( struct site_t *the_site )  {
    site_writefiles( the_site );
    return 0;
}

int site_list( struct site_t *the_site ) {
    struct site_file_t *current;
    int count;
    if( the_site->numnew > 0 ) {
	printf( "* These items have been added since the last update:\n" );
	count = 0;
	for( current = the_site->files; current!=NULL; current=current->next) {
	    if( current->diff == file_new ) {
		if( count++ ) printf( ", " );
		if( current->dir ) printf( "dir:" );
		printf( "%s", current->filename);
	    }
	}
	printf("\n");
    }
    if( the_site->numchanged > 0 ) {
	printf( "* These items have been changed since the last update:\n" );
	count = 0;
	for( current = the_site->files; current!=NULL; current=current->next) {
	    if( current->diff == file_changed ) {
		if( count++ ) printf( ", " );
		if( current->dir ) printf( "dir:" );
		printf( "%s", current->filename);
	    }
	}
	printf("\n");
    }
    if( the_site->numdeleted > 0 ) {
	printf( "* These items have been deleted since the last update:\n" );
	count = 0;
	for (current = the_site->files; current!=NULL; current=current->next) {
	    if( current->diff == file_deleted ) {
		if( count++ ) printf( ", " );
		if( current->dir ) printf( "dir:" );
		printf( "%s", current->filename);
	    }
	}
	printf("\n");
    }
    if( the_site->nummoved > 0 ) {
	printf( "* These items have been moved since the last update:\n" );
	count = 0;
	for (current = the_site->files; current!=NULL; current=current->next) {
	    if( current->diff == file_moved ) {
		if( count++ ) printf( ", " );
		if( current->dir ) printf( "dir:" ); /* later, later */
		printf( "%s->%s", current->old->filename, 
			current->filename );
	    }
	}
	printf("\n");
    }
    return 0;
}

/* Updates the remote site with the local copy.
 * This is the core functionality of sitecopy.
 * All we do is compare the local files list with the remote files
 * list, and upload any changed or new files. */
int site_update( struct site_t *the_site )  {
    int ret = 0;
    struct site_file_t *current;
    /* Connect to the server */
    printf( "Connecting to server... " );
    if( ftp_connect( the_site->server, 21 ) != FTP_OK ) {
	printf( "could not connect!\n%s\n", ftp_error );
	ftp_close( );
	return 2;
    }
    printf( "connected.\nLogging in... " );
    /* Now log in */
    if( ftp_login( the_site->username, the_site->password ) != FTP_OK ) {
	printf( "could not log in!:\n%s\n", ftp_error );
	ftp_close( );
	return 2;
    }
    printf( "logged in.\n" );
    
    for( current=the_site->files; current!=NULL; current=current->next ) {
	if( current->dir && (current->diff == file_new ) ) {
	    /* New dir! */
	    if( prompting ) {
		printf( "Create directory: %s", current->filename );
		if( !prompt() ) continue;
	    } else {
		printf( "Creating directory: %s\n", current->filename );
	    }
	    if( ftp_mkdir( current->remotename ) != FTP_OK ) {
		printf("Error! Could not create directory:\n%s\n", ftp_error );
		ret = 1;
	    } else {
		current->updated = true;
	    }
	}
    }

    for( current=the_site->files; current!=NULL; current=current->next ) {
	if( !current->dir ) {
	    switch( current->diff ) {
	    case file_new: /* File is new, upload it */
		if( prompting ) {
		    printf( "Upload new file: %s", current->filename );
		    if( !prompt() ) break;
		} else {
		    printf( "Uploading new file: %s\n", current->filename );
		}
		if( ftp_put( current->localname, current->remotename ) != FTP_OK ) {
		    printf( "Error! Could not upload file:\n%s\n", ftp_error );
		    ret = 1;
		} else { 
		    /* Successful upload! */
		    current->updated = true;
		    if( (the_site->perms == sitep_all) ||
			( (current->mode & S_IXUSR) &&
			  (the_site->perms == sitep_exec) ) ) {
			/* We need to chmod the file ... */
			if( prompting ) {
			    printf( "Set file permissions to %03o", current->mode & 0777 );
			    if( !prompt() ) break;
			} else {
			    printf( "Setting file permissions to %03o\n", current->mode & 0777);
			}
			if( ftp_chmod( current->remotename, current->mode ) != FTP_OK ) {
			    printf( "Error! Could not set permissions:\n%s\n", ftp_error );
			    ret = 1;
			}
		    }
		}
		break;

	    case file_changed: /* File has changed, upload it */
		if( prompting ) {
		    printf("Upload changed file: %s", current->filename );
		    if( !prompt() ) continue;
		} else {
		    printf( "Uploading changed file: %s\n", current->filename );
		}
		if( ftp_put( current->localname, current->remotename ) != FTP_OK ) {
		    printf( "Error! Could not upload file:\n%s\n", ftp_error );
		    ret = 1;
		} else {
		    /* Successful upload */
		    current->updated = true;
		}
		break;
		
	    case file_deleted: /* File has been removed, delete it */
		if( the_site->nodelete ) break; /* unless we should keep it */
		if( prompting ) {
		    printf( "Delete file: %s", current->filename );
		    if( !prompt() ) break;
		} else {
		    printf( "Deleting file: %s\n", current->filename );
		}
		if( ftp_delete( current->remotename ) != FTP_OK ) {
		    printf( "Error! Could not delete file:\n%s\n", ftp_error );
		    ret = 1;
		} else {
		    /* Successful update - file was deleted */
		    current->updated = true;
		}
		break;

	    case file_moved:
		/* The file has been moved */
		if( prompting ) {
		    printf( "Move file: %s -> %s", current->old->filename,
			current->filename );
		    if( !prompt() ) break;
		} else {
		    printf( "Moving file: %s -> %s\n", current->old->filename, current->filename);
		}
		if( ftp_move( current->old->remotename, 
			      current->remotename ) != FTP_OK ) {
		    ret = 1;
		    printf( "Error! Could not move file:\n%s\n", ftp_error );
		} else {
		    /* Successful update - file was moved */
		    current->updated = true;
		}		    
		break;

	    default: /* Ignore everything else */
		break;
	    }
	}
    }
    if( ! the_site->nodelete ) {
	for( current=the_site->files; current!=NULL; current=current->next ) {
	    if( current->dir && (current->diff == file_deleted ) ) {
		if( prompting ) {
		    printf( "Delete directory: %s", current->filename );
		    if( !prompt() ) continue;
		} else {
		    printf("Deleting directory: %s\n",current->filename );
		}
		if( ftp_rmdir( current->remotename ) != FTP_OK ) {
		    printf( "Error! Could not delete directory:\n%s\n", ftp_error);
		    ret = 1;
		} else {
		    /* Successful delete */
		    current->updated = true;
		}
	    }
	}
    }
 
    /* Close the connection */
    printf( "Closing connection... " );
    ftp_close();
    printf( "done.\n" );
    site_writefiles( the_site );
    return ret;
}

void site_assignnames( struct site_t *the_site, struct site_file_t *the_file ) {
    the_file->remotename = malloc( strlen(the_file->filename) +
				  strlen(the_site->remote_dir) + 1 );
    the_file->localname = malloc( strlen(the_file->filename) +
				 strlen(the_site->local_dir) + 1 );
    strcpy( the_file->remotename, the_site->remote_dir );
    strcat( the_file->remotename, the_file->filename + 1);
    strcpy( the_file->localname, the_site->local_dir );
    strcat( the_file->localname, the_file->filename + 1 );
}

/* This reads off the remote files and the local files */
int site_readfiles( struct site_t *the_site ) {
    FILE *fp;
    char buf[BUFSIZ], tmp[BUFSIZ];
    struct site_file_t *current, *previous;
    /* Get the remote files from the storage file */
    fp = fopen( the_site->infofile, "r" );
    if( fp != NULL ) {
	/* The file exists, so read it... */
	current = NULL;
	while( fgets( buf, BUFSIZ, fp) != NULL ) {
	    previous = current;
	    current = malloc( sizeof(struct site_file_t) );
	    if( previous ) {
		previous->next = current;
	    } else {
		the_site->files = current;
	    }
	    current->prev = previous;
	    current->filename = malloc( BUFSIZ );
	    sscanf( buf, "%s %s", current->filename, tmp );
	    site_assignnames( the_site, current );
	    current->diff = file_deleted;
	    the_site->numdeleted++;
	    if( debug_level==debug_full )
		fprintf( stderr, "Remote [%s]: ", current->filename );
	    if( strcmp( tmp, DIRWORD ) == 0 ) {
		/* It's a directory */
		current->dir = true;
		if( debug_level==debug_full )
		    fprintf( stderr, "directory.\n" );
	    } else {
		current->dir = false;
		current->remotetime = atol( tmp );
		sscanf( buf, "%s %s %ld", tmp,tmp, &current->remotesize );
		if( debug_level==debug_full ) {
		    if( debug_level==debug_full )
			fprintf( stderr, "file: size: %ld  time: %ld\n", current->remotesize, current->remotetime );
		}
	    }
	}
	fclose( fp );
    }
    /* Now read the local files */
    cur_site = the_site; /* So the callback knows what's going on */
    ftw( the_site->local_dir, ftw_callback, 10 );
    cur_site = NULL; /* To prevent me being stupid */
    /* Check for moved files - compare new files with deleted ones */
    if( the_site->checkmoved ) {
	site_checkmoved( the_site );
    }
    /* Summing up... */
    if( (the_site->numchanged +
	 the_site->nummoved + 
	 the_site->numnew +
	 (the_site->nodelete?0:the_site->numdeleted) ) > 0 ) {
	the_site->has_changed = true;
    } else {
	the_site->has_changed = false;
    }	
    return 0;
}

void site_checkmoved( struct site_t *the_site ) {
    /* Have any files been moved?
     * We look for new files, then see if we can find a deleted file
     * which matches filename, mod time, and size exactly.
     * 
     * This is simplistic, and it is possible to get it wrong in some
     * (unusual) circumstances:
     *    Old: /bar/file1.zip, /foo/file1.zip 
     *    New: /new/file1.zip.
     * If all the file1.zip's have the same size and mod time, it
     * IS NOT POSSIBLE to know which one has become /new/file1.zip without 
     * storing remembering the contents of /bar/file1.zip somehow and 
     * comparing, or using a filesystem which remembers stuff. In any case
     * it's pretty silly to try - so the whole shabang is optional.
     *
     * It would probably be better to search for DELETED files, then 
     * check them against NEW files, since files are more often added
     * to sites then they are deleted. But hey, this is C, who'll notice?
     */
    struct site_file_t *search, *current;
    char *bname, *tmp;
    /* First off, find the new files... */
    if( debug_level==debug_full )
	fprintf( stderr, "Checking for moved files...\n" );
    for( search=the_site->files; search!=NULL; search=search->next ) {
	if( search->diff != file_new )
	    continue;
	/* OK, we have us a new file... now iterate through the deleted
	 * files and look for a match */
	if( debug_level==debug_full )
	    fprintf( stderr, "- New file: %s\nComparing:", search->filename );
	bname = basename( search->filename );
	for( current=the_site->files; current!=NULL ; current=current->next ) {
	    /* Ignore files which aren't deleted, and directories completely.
	     * (search->diff==file_new, so we DON'T need to check
	     * that current != search as well) */
	    if( (current->diff != file_deleted) || current->dir ) {
		continue;
	    }
	    if( debug_level==debug_full )
		fprintf( stderr, " %s", current->filename );
	    tmp = basename( current->filename );
	    if( (strcmp( bname, tmp ) == 0) &&
		current->remotesize == search->localsize &&
		current->remotetime == search->localtime ) {
		if( debug_level==debug_full )
		    fprintf( stderr, " - matched!\n" );
		break;
	    }
	}
	if( current != NULL ) {
	    /* OK, current == the deleted file, search == the new file.
	     * First make sure search knows where it WAS before the move... */
	    search->old = current;
	    /* Remove the deleted file from the list so it doesn't
	     * actually get deleted by site_update */
	    LL_DELETE( current, the_site->files );
	    /* Diddle the totals a bit */
	    search->diff = file_moved;
	    the_site->numdeleted--;
	    the_site->numnew--;
	    the_site->nummoved++;
	} else if( debug_level==debug_full ) {
	    fprintf( stderr, " - No match found.\n" );
	}
    }
    if( debug_level==debug_full )
	fprintf( stderr, "Finished checking for moved files.\n" );
}

/* FTW on other 'nixes has const struct - not Linux. D'oh, I say */

#ifdef FTWCONST
int ftw_callback( const char *file, struct stat *sb, int flag ) {
#else
int ftw_callback( const char *file, const struct stat *sb, int flag ) {
#endif
    char *fname;
    struct site_file_t *current, *last;

    /* Get the filename relative to local root - chop the beginning
     * bit off */
    fname = (char *)file+strlen(cur_site->local_dir);

    if( strlen(fname) == 0) return 0;
    if( debug_level==debug_full )
	fprintf( stderr, "Local [%s]: ", fname );

    /* Now process it */
    last = NULL;
    for( current=cur_site->files; current!=NULL; current=current->next ) {
	if( strcmp( current->filename, fname ) == 0 )
	    /* Match found! */
	    break;
	last = current;
    }
    
    switch( flag ) {
    case FTW_F: /* File */
	if( debug_level==debug_full ) fprintf( stderr, "file - ");
	if( current == NULL ) {
	    /* This is a new local file... */
	    cur_site->numnew++;
	    current = malloc( sizeof(struct site_file_t) );
	    memset( current, 0, sizeof(struct site_file_t) );
	    if( (current->next = cur_site->files) )
		current->next->prev = current;
	    cur_site->files = current;
	    current->filename = strdup( fname ); 
	    site_assignnames( cur_site, current );
	    current->localtime = sb->st_mtime;
	    current->localsize = sb->st_size;
	    current->diff = file_new;
	    current->mode = sb->st_mode;
	    current->dir = false;
	    if( debug_level==debug_full )
		fprintf( stderr, "new; size: %ld, time: %ld\n", sb->st_size, sb->st_mtime ); 
	} else {
	    /* Match found... compare it */
	    current->localtime = sb->st_mtime;
	    current->localsize = sb->st_size;
	    cur_site->numdeleted--;
	    if( current->localtime > current->remotetime ) {
		current->diff = file_changed;
		cur_site->numchanged++;
		current->mode = sb->st_mode;
		if( debug_level==debug_full )
		    fprintf( stderr, "changed (%ld -> %ld)\n", current->localtime, current->remotetime );
	    } else {
		current->diff = file_unchanged;
		current->updated = true;
		cur_site->numunchanged++;
		if( debug_level==debug_full )
		    fprintf( stderr, "unchanged.\n" );
	    }
	} 
	break;
    case FTW_D:
	if( debug_level==debug_full )
	    fprintf( stderr, "directory - "); 	
	if( current == NULL ) {
	    /* New Directory */
	    cur_site->numnew++;
	    current = malloc( sizeof(struct site_file_t) );
	    memset( current, 0, sizeof(struct site_file_t) );
	    /* Directories go on the END of the list */
	    if( last ) {
		last->next = current;
	    } else {
		cur_site->files = current;
	    }
	    current->prev = last;
	    current->filename = strdup( fname );
	    site_assignnames( cur_site, current );
	    current->dir = true;
	    current->diff = file_new;
	    current->mode = sb->st_mode;
	    current->localsize = sb->st_size;
	    if( debug_level==debug_full )
		fprintf( stderr, "new.\n" ); 
	} else {
	    /* Existing directory */
	    cur_site->numdeleted--;
	    cur_site->numunchanged++;
	    current->diff = file_unchanged;
	    current->updated = true;
	    if( debug_level==debug_full )
		fprintf( stderr, "existing.\n" ); 
	}
    default:
	/* printf( "unknown!\n" ); */
	break;
    }
    /* carry on */
    return 0;
}

/* This writes the remote files list back to the file
 * and has some hairy logic in the middle, but is otherwise
 * pretty simple */
int site_writefiles( struct site_t *the_site ) {
    FILE *fp;
    struct site_file_t *current;
    fp = fopen( the_site->infofile, "w" );
    if( fp == NULL ) {
	/* Couldn't open! */
	perror( "fopen" );
	fprintf( stderr, "%s: Could not open site info file for writing: %s\n", progname, the_site->infofile );
	return 1;
    }
    for( current = the_site->files; current!=NULL; current=current->next ) {
	if( current->dir ) {
	    if( current->updated ) {
		switch( current->diff ) {
		case file_new:
		case file_unchanged:
		    /* It's new, or it's already there... */
		    fprintf( fp, "%s\t%s\n", current->filename, DIRWORD );
		    break;
		case file_deleted:
		default:
		    /* It's not there */
		    break;
		} 
	    } else {
		/* The update was not successful */
		switch( current->diff ) {
		case file_unchanged: /* ... it was there anyway */
		case file_deleted: /* ... but it wasn't deleted */
		    /* So the directory exists remotely */
		    fprintf( fp, "%s\t%s\n", current->filename, DIRWORD );
		    break;
		case file_new: /* ... but it wasn't created */
		default:
		    /* So the directory does not exist remotely */
		    break;
		}
	    }
	} else {
	    if( current->updated ) {
		switch( current->diff ) {
		case file_unchanged:
		    /* Write state of remote file */
		    fprintf( fp, "%s\t%ld\t%ld\n", current->filename, current->remotetime, current->remotesize );
		    break;
		case file_moved:
		case file_changed:
		case file_new:
		    /* Write state of local file */
		    fprintf( fp, "%s\t%ld\t%ld\n", current->filename, current->localtime, current->localsize );
		    break;
		case file_deleted:
		    /* So don't write anything out */
		    break;
		}
	    } else {
		switch( current->diff ) {
		case file_changed: /* ... but it hasn't been updated */
		case file_unchanged: 
		case file_deleted: /* ... but it hasn't been deleted */
		    /* Write out the remote time */
		    fprintf( fp, "%s\t%ld\t%ld\n", current->filename, current->remotetime, current->remotesize );
		    break;
		case file_moved: 
		    /* But it hasn't actually been moved, so write out
		     * the 'deleted' file (where the moved file WAS), and 
		     * ignore the 'new' file (where it was moved TO) */
		    fprintf( fp, "%s\t%ld\t%ld\n", current->old->filename, current->old->remotetime, current->old->remotesize );
		    break;
		case file_new: /* So ignore it... */
		    break;
		}
	    }
	}
    }
    fclose( fp );
    return 0;
}

int site_catchup( struct site_t *the_site ) {
    struct site_file_t *current;
    int ret;
    fprintf( stdout, "Catching up..." );
    fflush( stdout );
    for( current=the_site->files; current!=NULL; current=current->next ) {
	current->updated = true;
    }
    printf( "done.\n" );
    ret = site_writefiles( the_site );
    printf( "All files and directories are marked as updated remotely\n" );
    return ret;
}

void site_dumpsites( struct site_t *the_sites ) {
    struct site_t *current;
    for( current=the_sites; current!=NULL; current=current->next ) {
	printf( "\
Site: %s\n\
\tServer: %s   Username: %s\n\
\tRemote directory: %s%s\n\
\tLocal directory: %s\n\
\tPermissions: %s\n", 
	current->name, current->server, current->username, 
	DIRPREFIX(current->remote_dir),
	current->remote_dir, current->local_dir, (current->perms==sitep_ignore)?"Ignored":(current->perms==sitep_exec)?"Maintained for executables":"Always maintained");
	if( current->nodelete ) printf( "\tRemote files will not be deleted.\n" );
	if( current->checkmoved ) printf( "\tFiles will be moved remotely if moved locally.\n" );
    }
}
